#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import subprocess
import time
import datetime
import socket
import psutil
import uuid
import numpy as np

from sciveo.common.tools.logger import *
from sciveo.common.tools.daemon import DaemonBase
from sciveo.common.tools.hardware import *
from sciveo.common.tools.formating import format_memory_size
from sciveo.api.base import APIRemoteClient


class BaseMonitor(DaemonBase):
  def __init__(self, period=5):
    super().__init__(period=period)

    self.data = HardwareInfo()()
    self.data.setdefault("CPU", {})
    self.data.setdefault("DISK", {})
    self.data.setdefault("NET", {})
    self.data.setdefault("TEMP", {})
    self.data["RAM"] = {}
    self.data["LOG"] = {}
    self.list_logs = []

    self.api = APIRemoteClient()

    # Warmup the psutil
    psutil.cpu_percent(interval=0.3, percpu=True)
    initial_cpu_usage = psutil.cpu_percent(interval=None, percpu=True)
    self.previous_io_counters = {
      "DISK": psutil.disk_io_counters(perdisk=False),
      "NET": psutil.net_io_counters(pernic=False)
    }
    self.previous_time = {"DISK": time.time(), "NET": time.time()}
    time.sleep(1)

    machine_serial = self.getserial()

    debug(type(self).__name__, f"init monitor with period={period}", machine_serial, "initial_cpu_usage", initial_cpu_usage)

  def __call__(self):
    return self.data

  def loop(self):
    self.get_cpu_usage()
    self.get_temperatures()
    self.get_memory()
    self.get_gpu()
    self.get_disk()
    self.get_network()

    self.tail_logs()
    self.data["local_time"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    api_result = self.api.POST_SCI("monitor", {"data": self.data})

    debug(type(self).__name__, self(), "api_result", api_result)

  def tail_logs(self):
    for log_name, log_path in self.list_logs:
      self.data["LOG"][log_name] = self.tail_file(log_path)[-3:]

  def get_cpu_usage(self):
    try:
      usage_per_core = psutil.cpu_percent(interval=None, percpu=True)
      self.data["CPU"]["usage per core"] = usage_per_core
      self.data["CPU"]["usage"] = np.array(usage_per_core).mean()
    except Exception:
      pass

  def get_temperatures(self):
    try:
      temperatures = psutil.sensors_temperatures()
      for k, v in temperatures.items():
        self.data["TEMP"][k] = [t.current for t in v]
    except Exception:
      pass

  def get_memory(self):
    try:
      memory = psutil.virtual_memory()
      self.data["RAM"]["used"] = memory.used
      self.data["RAM"]["total"] = memory.total
      self.data["RAM"]["free"] = memory.free
      # self.data["RAM"]["installed"] = format_memory_size(memory.total)
      self.data["RAM"]["print"] = f"total: {format_memory_size(memory.total)} used: {format_memory_size(memory.used)}"
    except Exception:
      pass

  def getserial(self):
    machine_serial = None
    try:
      machine_serial = f"{socket.gethostname()}-{uuid.getnode()}"
    except Exception:
      pass
    if machine_serial is None:
      try:
        machine_serial = f"UUID-{uuid.getnode()}"
      except Exception:
        pass
    if machine_serial is None:
      machine_serial = f"RND-{random_token(8)}"
    self.data["serial"] = machine_serial
    return machine_serial

  def tail_file(self, file_path, block_size=1024):
    result = ["EMPTY"]
    try:
      with open(file_path,'rb') as fp:
        fp.seek(-block_size, os.SEEK_END)
        result = str(fp.read(block_size).rstrip()).split("\\n")
    except Exception as e:
      error(e, "tail_file", file_path)
    return result

  # Currently simple nvidia-smi wrapper impl
  def get_gpu(self):
    try:
      result = subprocess.run(
        [
          'nvidia-smi',
          '--query-gpu=gpu_uuid,gpu_name,index,power.draw,fan.speed,memory.total,memory.used,memory.free,utilization.gpu,utilization.memory,temperature.gpu',
          '--format=csv,nounits'
        ],
        capture_output=True, text=True, check=True
      )

      lines = result.stdout.strip().split('\n')
      self.data["GPU"] = {
        "raw_lines": lines
      }
    except Exception as e:
      pass

  def get_disk(self):
    try:
      list_metrics = ["read bytes", "write bytes", "read count", "write count", "read time", "write time"]
      disk_io_counters = psutil.disk_io_counters(perdisk=False)
      self.get_io_metrics("DISK", list_metrics, disk_io_counters)

      disk_usage = psutil.disk_usage('/')._asdict()
      self.data["DISK"]["print"] = f"{disk_usage['percent']}% ({round(disk_usage['used'] / (1024 * 1024 * 1024), 1)} GB / {round(disk_usage['total'] / (1024 * 1024 * 1024), 1)} GB)"
    except Exception as e:
      pass

  def get_network(self):
    try:
      list_metrics = ["bytes sent", "bytes recv", "packets sent", "packets recv"]
      net_io_counters = psutil.net_io_counters(pernic=False)
      self.get_io_metrics("NET", list_metrics, net_io_counters)
    except Exception as e:
      pass

  def get_io_metrics(self, name, list_metrics, io_counters):
    counters = io_counters._asdict()
    prev_io_counters = self.previous_io_counters[name]._asdict()
    time_diff = time.time() - self.previous_time[name]
    for metric_name in list_metrics:
      metric_key = metric_name.replace(' ', '_')
      self.data[name][metric_name] = counters[metric_key] - prev_io_counters[metric_key]

      if "bytes" in metric_name:
        speed_metric = metric_name.replace("bytes", "speed")
        self.data[name][speed_metric] = self.data[name][metric_name] / time_diff
    self.previous_time[name] = time.time()
    self.previous_io_counters[name] = io_counters

if __name__ == "__main__":
  mon = BaseMonitor(period=10)
  mon.start()

  while(True):
    time.sleep(30)